/*

linux/fs/dmsdos/dmsdos_ioctl.c

DMSDOS filesystem: special functions for further dmsdos tools

******************************************************************************
DMSDOS (Doublespace/Drivespace compressed MSDOS filesystem) for Linux
written 1995,1996 by Frank Gockel

    (C) Copyright 1995,1996 by Frank Gockel

Some code of the dmsdos filesystem has been copied from the msdos filesystem
so there are the following additional copyrights:

    (C) Copyright 1992,1993 by Werner Almesberger (msdos filesystem)
    (C) Copyright 1994,1995 by Jacques Gelinas (mmap code)
    (C) Copyright 1992-1995 by Linus Torvalds

The DMSDOS filesystem was inspired by the THS filesystem (a simple doublespace
DS-0-2 compressed read-only filesystem) written 1994 by Thomas Scheuermann.

The DMSDOS filesystem is distributed under the Gnu General Public Licence.
See file COPYING for details.
******************************************************************************

*/

#include <asm/segment.h>
#include <linux/errno.h>
#include <linux/kernel.h>
#include <linux/sched.h>
#include <linux/fs.h>
#include <linux/msdos_fs.h>
#include "dmsdos.h"

#ifdef INTERNAL_DAEMON
int idmsdosd(void*);
#define __KERNEL_SYSCALLS__
#include <linux/unistd.h>
#endif

extern unsigned long dmsdos_speedup;

int daemon_present=0;
int must_maintain_list=0;
int listcount=0;
Rwlist rwlist[LISTSIZE];
int rlist=0;
int wlist=0;

#ifdef INTERNAL_DAEMON
struct wait_queue * daemon_wait = NULL;
struct wait_queue * daemon_exit_wait=NULL;
int daemon_go_home=0;
int internal_daemon_counter=0;
#endif

void dumpcache(void);

struct wait_queue * list_wait = NULL;
int list_lock=0;
void lock_listaccess(void)
{ while(list_lock)sleep_on(&list_wait);
  list_lock=1;
}
void unlock_listaccess(void)
{ list_lock=0;
  wake_up(&list_wait);
}

void log_statistics(void)
{ log_list_statistics();
  log_ccache_statistics();
  log_found_statistics();
  /*log_other_statistics();*/
}

int set_maxcluster(struct super_block*sb,int data)
{ struct buffer_head*bh,*bh2;
  int i;
  unsigned char*pp;
  unsigned long int sectors;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;

  if(dblsb->s_cvf_version>=DRVSP3)
  { printk("DMSDOS: set_maxcluster refused - CVF is not in doublespace or drivespace<=2 format.\n");
    return -EINVAL;
  }
  if(data<2||data>dblsb->s_max_cluster2)return -EINVAL;
  /* check if higher clusters are unused */
  for(i=dblsb->s_max_cluster2;i>data;--i)
  { if(dbl_fat_nextcluster(sb,i,NULL))
    { printk("DMSDOS: set_maxcluster %d refused: cluster %d in use\n",
             data,i);
      return -EINVAL;
    }
  }
  
  bh=raw_bread(sb,0);
  if(bh==NULL)return -EIO;
  bh2=raw_bread(sb,dblsb->s_bootblock);
  if(bh2==NULL)
  { raw_brelse(sb,bh);
    return -EIO;
  }
  
  /* calculate number of sectors */
  /*sectors=(data-1)<<4;*/
  sectors=(data-1)*dblsb->s_sectperclust;
  sectors+=dblsb->s_rootdiranzentry>>4;
  pp=&(bh->b_data[14]);
  sectors+=CHS(pp);
  pp=&(bh->b_data[22]);
  sectors+=CHS(pp);
  bh->b_data[32]=sectors;
  bh->b_data[33]=sectors>>8;
  bh->b_data[34]=sectors>>16;
  bh->b_data[35]=sectors>>24;
  sectors+=CHS(pp);
  bh2->b_data[32]=sectors;
  bh2->b_data[33]=sectors>>8;
  bh2->b_data[34]=sectors>>16;
  bh2->b_data[35]=sectors>>24;
  raw_mark_buffer_dirty(sb,bh,1);
  raw_mark_buffer_dirty(sb,bh2,1);
  raw_brelse(sb,bh);
  raw_brelse(sb,bh2);
  
  dblsb->s_max_cluster=data;
  
  return 0;
}

void dmsdos_extra_statfs(struct super_block*sb, Dblstat*dblstat)
{ int sector,size,cluster;
  Mdfat_entry mde;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
    
      dblstat->free_sectors=0;
      dblstat->max_hole=0;  
      for(sector=dblsb->s_datastart;sector<dblsb->s_dataend;
          ++sector)
      { if(dbl_bitfat_value(sb,sector,NULL)==0)
        { ++(dblstat->free_sectors);
          size=1;
          while(dbl_bitfat_value(sb,sector+size,NULL)==0)++size;
          if(size>dblstat->max_hole)dblstat->max_hole=size;
          sector+=size-1;
          dblstat->free_sectors+=size-1;
        }
      }
      dblstat->used_sectors=dblsb->s_dataend+1-dblsb->s_datastart
                   -dblstat->free_sectors;
      dblstat->sectors_lo=0;
      dblstat->sectors_hi=0;
      dblstat->compressed_clusters=0;
      dblstat->uncompressed_clusters=0;
      dblstat->free_clusters=0;
      dblstat->used_clusters=0;
      dblstat->lost_clusters=0;
      for(cluster=2;cluster<=dblsb->s_max_cluster;++cluster)
      { if(dbl_fat_nextcluster(sb,cluster,NULL)!=0)
        { dbl_mdfat_value(sb,cluster,NULL,&mde);
          if(mde.flags&2)
          { ++(dblstat->used_clusters);
            if(mde.flags&1)++(dblstat->uncompressed_clusters);
            else ++(dblstat->compressed_clusters);
            dblstat->sectors_hi+=mde.size_hi_minus_1+1;
            dblstat->sectors_lo+=mde.size_lo_minus_1+1;
          }
          else ++(dblstat->lost_clusters);
        }
        else ++(dblstat->free_clusters);
      }
}
    
int dmsdos_ioctl_dir(
	struct inode *dir,
	struct file *filp,
	unsigned int cmd,
	unsigned long data)
{ unsigned char* idata;
  struct buffer_head*bh;
  int newval;
  int val;
  int cluster;
  int sector;
  Dblstat dblstat;
  Mdfat_entry mde,dummy;
  unsigned char *clusterd;
  int membytes;
  struct super_block*sb=dir->i_sb;
  Dblsb*dblsb=MSDOS_SB(sb)->private_data;
#ifndef INTERNAL_DAEMON
  int length;
  int i;
  int plist;
  /*int sectors;*/
  int rawlength;
#endif

  idata=(unsigned char*)data;

  switch(cmd)	
  { case DMSDOS_GET_DBLSB:
      /* check dutil version number */
      if(get_fs_long(idata)<DMSDOS_LOWEST_COMPATIBLE_VERSION)
        return DMSDOS_VERSION|0x0f000000;
      memcpy_tofs(idata,(unsigned char*)dblsb,sizeof(Dblsb));
      return DMSDOS_VERSION;
    case DMSDOS_EXTRA_STATFS:
      dmsdos_extra_statfs(dir->i_sb,&dblstat);
      memcpy_tofs(idata,&dblstat,sizeof(Dblstat));
      return 0;
    case DMSDOS_READ_BLOCK:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb->s_dataend)return -EINVAL;
      bh=raw_bread(dir->i_sb,sector);
      if(bh==NULL)return -EIO;
      memcpy_tofs(idata+sizeof(long),bh->b_data,SECTOR_SIZE);
      raw_brelse(dir->i_sb,bh);
      return 0;
    case DMSDOS_WRITE_BLOCK:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb->s_dataend)return -EINVAL;
      bh=raw_bread(dir->i_sb,sector);
      if(bh==NULL)return -EIO;
      memcpy_fromfs(bh->b_data,idata+sizeof(long),SECTOR_SIZE);
      raw_mark_buffer_dirty(dir->i_sb,bh,1);
      raw_brelse(dir->i_sb,bh);
      return 0;
    case DMSDOS_READ_DIRENTRY:
    case DMSDOS_WRITE_DIRENTRY:
      printk("DMSDOS: READ/WRITE DIRENTRY ioctl has gone\n");
      return -EINVAL;
    case DMSDOS_READ_BITFAT:
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb->s_dataend)return -EINVAL;
      val=dbl_bitfat_value(sb,sector,NULL);
      put_fs_long(val,idata+sizeof(long));
      return 0;
    case DMSDOS_WRITE_BITFAT:
      if(current->euid!=0)return -EPERM;
      sector=get_fs_long(idata);
      if(sector<0||sector>dblsb->s_dataend)return -EINVAL;
      newval=get_fs_long(idata+sizeof(long));
      dbl_bitfat_value(sb,sector,&newval);
      return 0;
    case DMSDOS_READ_MDFAT:
      cluster=get_fs_long(idata);
      if(cluster>dblsb->s_max_cluster)return -EINVAL;
      dbl_mdfat_value(sb,cluster,NULL,&mde);
      /*put_fs_long(val,idata+4);*/
      memcpy_tofs((Mdfat_entry*)get_fs_long(idata+sizeof(long)),&mde,
                  sizeof(Mdfat_entry));
      return 0;
    case DMSDOS_WRITE_MDFAT:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb->s_max_cluster)return -EINVAL;
      /*newval=get_fs_long(idata+4);*/
      memcpy_fromfs(&mde,(Mdfat_entry*)get_fs_long(idata+sizeof(long)),
                    sizeof(Mdfat_entry));
      dbl_mdfat_value(sb,cluster,&mde,&dummy);
      return 0;
    case DMSDOS_READ_DFAT:
      cluster=get_fs_long(idata);
      if(cluster>dblsb->s_max_cluster)return -EINVAL;
      val=dbl_fat_nextcluster(sb,cluster,NULL);
      put_fs_long(val,idata+sizeof(long));
      return 0;
    case DMSDOS_WRITE_DFAT:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster>dblsb->s_max_cluster)return -EINVAL;
      newval=get_fs_long(idata+sizeof(long));
      dbl_fat_nextcluster(sb,cluster,&newval);
      return 0;
    case DMSDOS_READ_CLUSTER:
      if(current->euid!=0)return -EPERM;
      cluster=get_fs_long(idata);
      if(cluster < 0 || cluster>dblsb->s_max_cluster)
        return -EINVAL;
      membytes=SECTOR_SIZE*dblsb->s_sectperclust;
      if((clusterd=(unsigned char*)MALLOC(membytes))==NULL)
      { printk("DMSDOS: ioctl: read_cluster: no memory!\n");
        return -EIO;
      }
      val=dmsdos_read_cluster(dir->i_sb,clusterd,cluster);
      if (val >= 0)
        memcpy_tofs(idata+sizeof(long), clusterd, membytes);
      FREE(clusterd);
      return val;
    case DMSDOS_SET_COMP:
      if(current->euid!=0)return -EPERM;
      dblsb->s_comp=data;
      return 0;
    case DMSDOS_SET_CF:
      if(current->euid!=0)return -EPERM;
      if(data>=12u)return -EINVAL;
      dblsb->s_cfaktor=data;
      return 0;
    case DMSDOS_SIMPLE_CHECK:
      val=simple_check(dir->i_sb,get_fs_long(idata));
      put_fs_long(val,idata);
      return 0;
    case DMSDOS_DUMPCACHE:
      dumpcache();
      return 0;
    case DMSDOS_D_ASK:
 #ifdef INTERNAL_DAEMON
      return -EINVAL;
 #else
      if(current->euid!=0)return -EPERM;
      /* meldet dmsdosd an */
      daemon_present=1;
      LOG_DAEMON("DMSDOS: D_ASK\n");
      return 0;
 #endif
    case DMSDOS_D_READ:
 #ifdef INTERNAL_DAEMON
      return -EINVAL;
 #else
      if(current->euid!=0)return -EPERM;
      lock_listaccess();
      /*search next valid entry*/
      for(i=LISTSIZE;i>0;--i)
      { if(rwlist[rlist].flag==D_VALID)
        { /* check in mdfat that cluster is actually used */
          dbl_mdfat_value(rwlist[rlist].sb,rwlist[rlist].clusternr,
                          NULL,&mde);
          if((mde.flags&3)==3)goto vr_found; /* used and uncompressed */
          rwlist[rlist].flag=D_EMPTY;  /* remove - it's garbage */
          --listcount;
          LOG_DAEMON("DMSDOS: D_READ: removing garbage entry cluster=%d\n",
                 rwlist[rlist].clusternr);
        }
        rlist=(rlist+1)&(LISTSIZE-1);
      }
      unlock_listaccess();
      return 0;
   vr_found:
      cluster=rwlist[rlist].clusternr;
      /* yes we change sb here */
      sb=rwlist[rlist].sb;
      dblsb=MSDOS_SB(sb)->private_data;
      length=rwlist[rlist].length;
      clusterd=MALLOC(dblsb->s_sectperclust*SECTOR_SIZE);
      if(clusterd==NULL)
      { printk("DMSDOS: ioctl: D_READ: no memory!\n");
        unlock_listaccess();
        return 0; 
      }
      if((i=dmsdos_read_cluster(sb,clusterd,cluster))<0)
      { printk("DMSDOS: ioctl: D_READ: read_cluster failed!\n");
        rwlist[rlist].flag=D_EMPTY;
        --listcount;
        unlock_listaccess();
        FREE(clusterd);
        return 0;   
      }
      if(length<0)length=i; /* if invalid length, use the read value */
      rwlist[rlist].flag=D_IN_D_ACTION;
      put_fs_long(rlist,idata);
      put_fs_long(length,idata+sizeof(long));
      put_fs_long(rwlist[rlist].method,idata+2*sizeof(long));
      memcpy_tofs(idata+3*sizeof(long),clusterd,length);
      unlock_listaccess();
      FREE(clusterd);
      return 1;
#endif
    case DMSDOS_D_WRITE:
#ifdef INTERNAL_DAEMON
      return -EINVAL;
#else
      if(current->euid!=0)return -EPERM;
      lock_listaccess();
      plist=get_fs_long(idata);
      if(rwlist[plist].flag==D_OVERWRITTEN){rwlist[plist].flag=D_EMPTY;--listcount;}
      if(rwlist[plist].flag!=D_IN_D_ACTION)
      { LOG_DAEMON("DMSDOS: D_WRITE: Entry not in action, flag=%d cluster=%d\n",
               rwlist[plist].flag,rwlist[plist].clusternr);
        unlock_listaccess();
        return 0;
      }
      
      /* will be freed in any case later, so: */
      --listcount;
      
      rawlength=get_fs_long(idata+sizeof(long));
      if(rawlength==0)
      { /* data were uncompressible */
        rwlist[plist].flag=D_EMPTY;
        unlock_listaccess();
        return 0;
      }
      /* check that cluster is used */
      dbl_mdfat_value(rwlist[plist].sb,rwlist[plist].clusternr,
                      NULL,&mde);
      if((mde.flags&3)!=3)
      { rwlist[plist].flag=D_EMPTY;  /* remove - it's garbage */
        LOG_DAEMON("DMSDOS: D_WRITE: removing garbage entry cluster=%d\n",
               rwlist[plist].clusternr);
        unlock_listaccess();
        return 0;
      }
      clusterd=MALLOC(rawlength);
      if(clusterd==NULL)
      { printk("DMSDOS: ioctl: D_WRITE: no memory!\n");
        rwlist[plist].flag=D_EMPTY;
        unlock_listaccess();
        return 0;
      }
      length=rwlist[plist].length;
      cluster=rwlist[plist].clusternr;
      sb=rwlist[plist].sb;
      memcpy_fromfs(clusterd,idata+3*sizeof(long),rawlength);
      dmsdos_write_cluster(sb,clusterd,length,
                           cluster,0,
                           -rawlength); /* ucflag < 0 means raw write */
      rwlist[plist].flag=D_EMPTY;
      unlock_listaccess();
      FREE(clusterd);
      return 0;
#endif
    case DMSDOS_D_EXIT:
#ifdef INTERNAL_DAEMON
      return -EINVAL;
#else
      if(current->euid!=0)return -EPERM;
      /* meldet dmsdosd ab */
      daemon_present=0;
      LOG_DAEMON("DMSDOS: D_EXIT\n");
      return 0;
#endif
    case DMSDOS_MOVEBACK:
      printk("DMSDOS: MOVEBACK ioctl has gone\n");
      return -EINVAL;
    case DMSDOS_SET_MAXCLUSTER:
      if(current->euid!=0)return -EPERM;
      return set_maxcluster(dir->i_sb,data);
      return -EINVAL;
    case DMSDOS_FREE_IDLE_CACHE:
      if(current->euid!=0)return -EPERM;
      free_idle_cache();
      return 0;
    case DMSDOS_SET_LOGLEVEL:
      if(current->euid!=0)return -EPERM;
      loglevel=data;
      printk("DMSDOS: ioctl: loglevel set to 0x%lx.\n",loglevel);
      return 0;
    case DMSDOS_SET_SPEEDUP:
      if(current->euid!=0)return -EPERM;
      dmsdos_speedup=data;
      printk("DMSDOS: ioctl: speedup set to 0x%lx.\n",dmsdos_speedup);
      return 0;
    case DMSDOS_SYNC_CCACHE:
      sync_cluster_cache(data);
      return 0;
    case DMSDOS_LOG_STATISTICS:
      if(current->euid!=0)return -EPERM;
      log_statistics();
      return 0;
    case DMSDOS_RECOMPRESS:
      printk("DMSDOS: RECOMPRESS ioctl has gone\n");
      return -EINVAL;
    default:
      return -EINVAL;
  }	
	
}

void remove_from_daemon_list(struct super_block*sb,int clusternr)
{ int i;

  if(must_maintain_list==0)return;

  lock_listaccess();
  
  /* check list for existing entry and mark it as overwritten */
  for(i=0;i<LISTSIZE;++i)
  { if(rwlist[i].flag!=D_EMPTY)
    if(rwlist[i].clusternr==clusternr&&rwlist[i].sb->s_dev==sb->s_dev)
    {
      if(rwlist[i].flag==D_IN_D_ACTION){rwlist[i].flag=D_OVERWRITTEN;break;}
      if(rwlist[i].flag==D_VALID){rwlist[i].flag=D_EMPTY;--listcount;break;}
    }
  }
  
  if(daemon_present==0)
  { if(listcount==0)must_maintain_list=0;
  }
    
  unlock_listaccess();
}

int try_daemon(struct super_block*sb,int clusternr, int length, int method)
{ int i;

  if(daemon_present==0&&must_maintain_list==0)return 0;
  
  lock_listaccess();
  
  must_maintain_list=1;
  
  /* check list for existing entry and mark it as overwritten */
/* no longer necessary here......
  for(i=0;i<LISTSIZE;++i)
  { if(rwlist[i].clusternr==clusternr&&rwlist[i].sb->s_dev==sb->s_dev)
    {
      if(rwlist[i].flag==D_IN_D_ACTION){rwlist[i].flag=D_OVERWRITTEN;break;}
      if(rwlist[i].flag==D_VALID){rwlist[i].flag=D_EMPTY;--listcount;break;}
    }
  }
*/
  
  if(daemon_present==0||listcount==LISTSIZE)
  {
    if(listcount==0)must_maintain_list=0;
    unlock_listaccess();
    return 0;
  }
 
  /* find empty slot in list */
  for(i=LISTSIZE;i>0;--i)
  { if(rwlist[wlist].flag==D_EMPTY) goto w_found;
    wlist=(wlist+1)&(LISTSIZE-1);
  }
  /*this shouldn't happen - otherwise there's a count error */
  printk("DMSDOS: try_daemon: no empty slot found, listcount corrected.\n");
  listcount=LISTSIZE;
  unlock_listaccess();
  return 0;
  
 w_found:
  ++listcount;
  rwlist[wlist].clusternr=clusternr;
  rwlist[wlist].sb=sb;
  rwlist[wlist].length=length;
  rwlist[wlist].method=method;
  rwlist[wlist].flag=D_VALID;
  unlock_listaccess();
#ifdef INTERNAL_DAEMON
  wake_up(&daemon_wait);
#endif
  return 1;
}

void clear_list_dev(struct super_block*sb)
{ int i;

  lock_listaccess();
  for(i=0;i<LISTSIZE;++i)
  { if(rwlist[i].flag!=D_EMPTY)
    { if(rwlist[i].sb)printk("DMSDOS: clear_list_dev: Uhh, sb==NULL ...\n");
      else
      if(rwlist[i].sb->s_dev==sb->s_dev)
      { rwlist[i].flag=D_EMPTY;
        --listcount;
      }
    }
  }
  unlock_listaccess();
}

void init_daemon(void)
{ int i;

  if(daemon_present)
  { printk("DMSDOS: init_daemon: daemon already present\n");
#ifdef INTERNAL_DAEMON
    ++internal_daemon_counter;
#endif
    return;
  }
  lock_listaccess();
  printk("DMSDOS: clearing rwlist...\n");
  for(i=0;i<LISTSIZE;++i)rwlist[i].flag=D_EMPTY;
  listcount=0;
  unlock_listaccess();

#ifdef INTERNAL_DAEMON
  /*daemon_present=1...this is maintained by idmsdosd itself*/;
  internal_daemon_counter=1;
  /* fire up internal saemon */
  printk("DMSDOS: starting internal daemon...\n");
  kernel_thread(idmsdosd,NULL,0);
#endif
}

void log_list_statistics()
{ int j;
  int empty;
  int valid;
  int in_action;
  int overwritten;
  
  lock_listaccess();
  
  printk("DMSDOS: list statistics:\n");
  printk("daemon_present=%d must_maintain_list=%d listcount=%d\n",
         daemon_present,must_maintain_list,listcount);
  empty=0;
  valid=0;
  in_action=0;
  overwritten=0;
    
  for(j=0;j<LISTSIZE;++j)
      { switch(rwlist[j].flag)
        { case D_EMPTY: ++empty; break;
          case D_VALID: ++valid; break;
          case D_IN_D_ACTION: ++in_action; break;
          case D_OVERWRITTEN: ++overwritten; break;
          default: printk("DMSDOS: log_list_statistics: cannot happen.\n");
        }
      }
  printk("sum: empty=%d valid=%d in_action=%d overwritten=%d\n",
         empty,valid,in_action,overwritten);
  
  unlock_listaccess();
}

#ifdef INTERNAL_DAEMON

int internal_d_read(int*val1, int*val2, int*val3,
                    unsigned char*clusterd)
{ int i;
  int cluster;
  Mdfat_entry mde;
  int length;
  int method;
  struct super_block*sb=NULL;

      lock_listaccess();
      /*search next valid entry*/
      for(i=LISTSIZE;i>0;--i)
      { if(rwlist[rlist].flag==D_VALID)
        { /* check in mdfat that cluster is actually used */
          sb=rwlist[rlist].sb;
          if(sb)
          { dbl_mdfat_value(sb,rwlist[rlist].clusternr,
                          NULL,&mde);
            if((mde.flags&3)==3)goto vr_found; /* used and uncompressed */
          }
          rwlist[rlist].flag=D_EMPTY;  /* remove - it's garbage */
          --listcount;
          LOG_DAEMON("DMSDOS: D_READ: removing garbage entry cluster=%d\n",
                 rwlist[rlist].clusternr);
        }
        rlist=(rlist+1)&(LISTSIZE-1);
      }
      unlock_listaccess();
      return 0;
   vr_found:
      cluster=rwlist[rlist].clusternr;
      sb=rwlist[rlist].sb;
      length=rwlist[rlist].length;
      method=rwlist[rlist].method;
      if((i=dmsdos_read_cluster(sb,clusterd,cluster))<0)
      { printk("DMSDOS: ioctl: D_READ: read_cluster failed!\n");
        rwlist[rlist].flag=D_EMPTY;
        --listcount;
        unlock_listaccess();
        return 0;   
      }
      if(length<0)length=i; /* if invalid length, use read value */
      rwlist[rlist].flag=D_IN_D_ACTION;
      *val1=rlist; /*put_fs_long(rlist,idata);*/
      *val2=length; /*put_fs_long(length,idata+sizeof(long));*/
      *val3=method; /*put_fs_long(rwlist[rlist].method,idata+2*sizeof(long));*/
      unlock_listaccess();
      return 1;
}

int internal_d_write(int plist, int rawlength, unsigned char*clusterd)
{ Mdfat_entry mde;
  int length;
  int cluster;
  struct super_block*sb=NULL;

      lock_listaccess();
      if(rwlist[plist].flag==D_OVERWRITTEN){rwlist[plist].flag=D_EMPTY;--listcount;}
      if(rwlist[plist].flag!=D_IN_D_ACTION)
      { LOG_DAEMON("DMSDOS: D_WRITE: Entry not in action, flag=%d cluster=%d\n",
               rwlist[plist].flag,rwlist[plist].clusternr);
        unlock_listaccess();
        return 0;
      }
      
      /* will be freed in any case later, so: */
      --listcount;
      
      if(rawlength==0)
      { /* data were uncompressible */
        rwlist[plist].flag=D_EMPTY;
        unlock_listaccess();
        return 0;
      }
      /* check that cluster is used */
      sb=rwlist[plist].sb;
      if(sb==NULL)goto shitt;
      dbl_mdfat_value(sb,rwlist[plist].clusternr,
                      NULL,&mde);
      if((mde.flags&3)!=3)
      { shitt:
        rwlist[plist].flag=D_EMPTY;  /* remove - it's garbage */
        LOG_DAEMON("DMSDOS: D_WRITE: removing garbage entry cluster=%d\n",
               rwlist[plist].clusternr);
        unlock_listaccess();
        return 0;
      }
      length=rwlist[plist].length;
      cluster=rwlist[plist].clusternr;
      dmsdos_write_cluster(sb,clusterd,length,cluster,0,
                           -rawlength); /* ucflag < 0 means raw write */
      rwlist[plist].flag=D_EMPTY;
      unlock_listaccess();
      return 0;
}

typedef struct
{ long val1;
  long val2;
  long val3;
  unsigned char data[32*1024];
} Cdata;

/* we need the memory always - there's only one process using it - idmsdosd */
Cdata cdata;
Cdata ckdata;

int get_and_compress_one(void)
{ int ret;
  int handle;
  int length;
  int size;
  int method;
  
  /* get cluster to compress */
  LOG_DAEMON("idmsdosd: Trying to read...\n");
  ret=internal_d_read(&handle,&length,&method,cdata.data);
  if(ret!=1)
  { LOG_DAEMON("idmsdosd: nothing there - D_READ ioctl returned %d\n",ret);
    return ret;
  }
  
  size=(length-1)/512+1;
  LOG_DAEMON("idmsdosd: compressing...\n");
  ret=(method==SD_3||method==SD_4) ?
       stac_compress(cdata.data,length,ckdata.data,
                     sizeof(ckdata.data),method,11) :
       dbl_compress(ckdata.data,cdata.data,size,method,11)*512;
  LOG_DAEMON("idmsdosd: compress %X from %d returned %d\n",
                  method,length,ret);
  if(ret<0)ret=0; /* compression failed */
  LOG_DAEMON("idmsdosd: writing...\n");
  internal_d_write(handle,ret,ckdata.data);
  
  return 1; /* one cluster compressed */
}

struct timer_list idmsdosd_timer;

static void idmsdosd_timer_function(unsigned long data)
{ 
  /* do something */
  /*printk("DMSDOS: idmsdosd_timer_function: doing something :)\n");*/
    
  /* wake up daemon */
  wake_up(&daemon_wait);

  del_timer(&idmsdosd_timer);
  idmsdosd_timer.expires=jiffies + (IDMSDOSD_TIME * HZ);
  add_timer(&idmsdosd_timer);
}

int idmsdosd(void*dummy)
{      
        /* throw away some things from the calling process */
        /* otherwise the root fs cannot be unmounted on reboot ... urgh*/
	exit_files(current);
	exit_fs(current);
	exit_sighand(current);
        exit_mm(current);
        
        /*
         *      We have a bare-bones task_struct, and really should fill
         *      in a few more things so "top" and /proc/2/{exe,root,cwd}
         *      display semi-sane things. Not real crucial though...
         */
        current->session = 1;
        current->pgrp = 1;
        sprintf(current->comm, "dmsdosd");

        /*
         *      As a kernel thread we want to tamper with system buffers
         *      and other internals and thus be subject to the SMP locking
         *      rules. (On a uniprocessor box this does nothing).
         */
#ifdef __SMP__
        lock_kernel();
        syscall_count++;
#endif

  daemon_present=1;
  
  init_timer(&idmsdosd_timer);
  idmsdosd_timer.function=idmsdosd_timer_function;
  idmsdosd_timer.expires=jiffies + (IDMSDOSD_TIME * HZ);
  add_timer(&idmsdosd_timer);
  
  for(;;)
  { while(get_and_compress_one()==1);
    /* don't kill the system performance when nothing to compress */
    { LOG_DAEMON("idmsdosd: sleeping...\n"); 
      interruptible_sleep_on(&daemon_wait);
      if(daemon_go_home)break;
      /* throw away long idle mdfat/dfat/bitfat sectors and clusters */
      free_idle_cache();
    }
  }
  
  del_timer(&idmsdosd_timer);
  daemon_present=0;
  
  LOG_DAEMON("idmsdosd: exiting...\n");
  wake_up(&daemon_exit_wait);
  return 0;
}

void remove_internal_daemon(void)
{ if(daemon_present)
   { printk("DMSDOS: killing internal daemon...\n");
     daemon_go_home=1;
     wake_up(&daemon_wait);
     interruptible_sleep_on(&daemon_exit_wait);
  /* hmm... don't know if this is enough... maybe we have to wait(syscall)
     for the daemon to exit and remove its task structure... 
     otherwise a zombie may be left behind after unloading the module */
   }
}

#endif

void exit_daemon(void)
{
#ifdef INTERNAL_DAEMON
   --internal_daemon_counter;
   if(internal_daemon_counter<0)
   { printk("DMSDOS: exit_daemon: counter<0 ???\n");
     internal_daemon_counter=0;
   }
   if(internal_daemon_counter==0)remove_internal_daemon();
#endif
}

void force_exit_daemon(void)
{
#ifdef INTERNAL_DAEMON
   internal_daemon_counter=0;
   remove_internal_daemon();
#endif
}
